library fields.field;

import 'dart:async';
import 'dart:mirrors';
import 'field_list.dart';
import 'mirrors.dart';

/// A marker class for field annotations.
abstract class FieldAnnotation {
  const FieldAnnotation();  
}

const IgnoreField = const _IgnoreField();
class _IgnoreField extends FieldAnnotation {
  const _IgnoreField();
}

class DisplayName extends FieldAnnotation {
  final String name;
  const DisplayName(this.name);
}

const ReadOnly = const _ReadOnly();
class _ReadOnly extends FieldAnnotation {
  const _ReadOnly();
}

class UseFieldEditor extends FieldAnnotation {
  final Type editorType;
  const UseFieldEditor(this.editorType);
}

class Field {
  
  final Symbol name;
  String _displayName;
  String get displayName => _displayName;
  final Type type;
  final bool isPrivate;
  final bool isReadOnly;
  final Field parent;
  final ClassMirror editor;
  final Iterable<FieldAnnotation> annotations;
  FieldList _children;
  FieldList get children => _children;
  InstanceMirror owner;  
  Object _value;
  
  final _values = new StreamController<Object>.broadcast();
  Stream<Object> get values => _values.stream;
  
  Field(this.owner, 
        this.name, 
        this.type, 
        {this.parent, 
        this.editor, 
        this.annotations: const [], 
        DisplayName displayName, 
        this.isPrivate: false, 
        this.isReadOnly: false}) {
    if (displayName == null) _displayName = MirrorSystem.getName(name);
    else _displayName = displayName.name;
    _children = new FieldList(this);
    _value = owner.getField(name).reflectee;
    if (!isSimpleType(type)) _children.refreshSource(_value);
  }
  
  Future<Object> getValue() {
    return owner.getFieldAsync(name)
        .then((instance) {
          if (instance.reflectee != _value) {
            _value = instance.reflectee;
            _refreshChildren().then((_) => _values.add(instance.reflectee));
          }
          return new Future.value(_value);
        })
        .catchError((e) => null,
              test: (e) => e is MirroredUncaughtExceptionError 
                        || e is MirroredCompilationError);
  }
  
  void refreshValue() { getValue(); }
  
  Future<Object> setValue(Object val) {
    if(isReadOnly) {
      return new Future.error(
          new StateError('Field $displayName is read-only.'));
    }
    else {
      if (val == _value) return new Future.value(val);
      else {
        return owner.setFieldAsync(name, reflect(val)).then((_) {
          _value = val;
          return _refreshChildren().then((_) {
            _values.add(val);
            if (parent != null) parent._values.add(parent._value);
            return new Future.value(val);
          });
        });
      }
    }
  }
  
  Future _refreshChildren() {    
    if (children.isEmpty) return new Future.value();
    children.refreshSource(_value);
    return Future.forEach(children, (f) => f.refreshValue());
  }
}
