library fields.field_transform;

/// Check if a transform from the [sourceType] to the [targetType] can be 
/// performed.  Returns _true_ if such a transform is possible, else _false_.
bool canTransformField(Type sourceType, Type targetType) {
  if (sourceType == targetType) return true;
  return FieldTransform._transforms.any(
      (t) => t.sourceType == sourceType && t.targetType == targetType);
}

/// Transform the given [value] from its [sourceType] to a [targetType].
/// Returns the transformed value.
transformField(value, Type sourceType, Type targetType) {
  if (sourceType == targetType) return value;
  return FieldTransform._transforms.firstWhere(
      (t) => t.sourceType == sourceType && t.targetType == targetType)(value);
}

/// An error encountered during a transform.
class FieldTransformError extends Error {
  /// The caught object that caused the transform to err; may be `null` if none.
  final cause;
  /// A description of this transform error.
  final String message;
  /// Constructs a new transform error with the optional [cause] and [message].
  FieldTransformError({this.cause: null, this.message: ''});
  String toString() => 'TransformError: $message |> $cause';
}

/// A one-directional transform from type [S] to type [T].
abstract class FieldTransform<S, T> {
  static final Set<FieldTransform> _transforms = new Set<FieldTransform>();
  
  static void register(FieldTransform transform) => _transforms.add(transform);
  
  /// Unregister the given [transform].
  /// Returns _true_ if the [transform] was registered, else returns _false_.
  static bool unregister(FieldTransform transform) => 
      _transforms.remove(transform);
  
  static void unregisterAll() => _transforms.clear();
  
  /// The source type of this transform.
  final Type sourceType = S;
  /// The target type of this transform.
  final Type targetType = T;
  const FieldTransform();
  /// Call this transform on [value].  Returns the transformed value.
  T call(S value);
}
