
library fields.test.field;

import 'dart:async';
import 'dart:mirrors';
import 'package:bench/bench.dart';
import 'package:unittest/unittest.dart';
import '../lib/src/field.dart';

main() => reflectTests();

// -----------------------------------------------------------------------------
// Setup
// -----------------------------------------------------------------------------

class CustomType {
  int i;
  int get ci => i;
      set ci(int v) => i = v;
  String s;
  
  String toString() => 'custom($i,$ci,$s)';
}

class HasFields {
  int i = 7;
  final int ri = 42;
  num n = 3.14;
  final num rn = 6.28;
  double d = 1024.4201;
  final double rd = 1.024;
  String s = 'snarf';
  final rs = 'immutable';
  List<int> li = [ 1, 2, 3, 4];  
  Map m = { 'a' : 1, 'b' : 2 };
  CustomType c = new CustomType();
}

// -----------------------------------------------------------------------------
// Basic
// -----------------------------------------------------------------------------

@Test('Test the construction of an int field.')
void testInt() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('i'), int);    
  f.getValue().then(expectAsync1((v) {
    expect(v, equals(7));    
    expect(f.name, equals(const Symbol('i')));
    expect(f.type, equals(int));
    expect(f.isReadOnly, isFalse);
    expect(f.children, isEmpty);
  }));
}

@Test('Test the construction of a read-only int field.')
void testIntReadOnly() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('ri'), int, isReadOnly: true);    
  f.getValue().then(expectAsync1((v) {
    expect(v, equals(42));    
    expect(f.name, equals(const Symbol('ri')));
    expect(f.type, equals(int));
    expect(f.isReadOnly, isTrue);
    expect(f.children, isEmpty);
  }));
}

@Test('Test the construction of a num field.')
void testNum() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('n'), num);    
  f.getValue().then(expectAsync1((v) {
    expect(v, equals(3.14));  
    expect(f.name, equals(const Symbol('n')));
    expect(f.type, equals(num));
    expect(f.isReadOnly, isFalse);
    expect(f.children, isEmpty);
  }));
}

@Test('Test the construction of a read-only num field.')
void testNumReadOnly() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('rn'), num, isReadOnly: true);    
  f.getValue().then(expectAsync1((v) {
    expect(v, equals(6.28));  
    expect(f.name, equals(const Symbol('rn')));
    expect(f.type, equals(num));
    expect(f.isReadOnly, isTrue);
    expect(f.children, isEmpty);
  }));
}

@Test('Test the construction of a double field.')
void testDouble() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('d'), double);    
  f.getValue().then(expectAsync1((v) {
    expect(v, equals(1024.4201));  
    expect(f.name, equals(const Symbol('d')));
    expect(f.type, equals(double));
    expect(f.isReadOnly, isFalse);
    expect(f.children, isEmpty);
  }));
}

@Test('Test the construction of a read-only double field.')
void testDoubleReadOnly() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('rd'), double, isReadOnly: true);    
  f.getValue().then(expectAsync1((v) {
    expect(v, equals(1.024));  
    expect(f.name, equals(const Symbol('rd')));
    expect(f.type, equals(double));
    expect(f.isReadOnly, isTrue);
    expect(f.children, isEmpty);
  }));
}

@Test('Test the construction of a string field.')
void testString() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('s'), String);    
  f.getValue().then(expectAsync1((v) {
    expect(v, equals('snarf'));  
    expect(f.name, equals(const Symbol('s')));
    expect(f.type, equals(String));
    expect(f.isReadOnly, isFalse);
    expect(f.children, isEmpty);
  }));
}

@Test('Test the construction of a read-only string field.')
void testStringReadOnly() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('rs'), String, isReadOnly: true);    
  f.getValue().then(expectAsync1((v) {
    expect(v, equals('immutable'));  
    expect(f.name, equals(const Symbol('rs')));
    expect(f.type, equals(String));
    expect(f.isReadOnly, isTrue);
    expect(f.children, isEmpty);
  }));
}

@Test('Test the construction of a list field.')
void testList() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('li'), List);    
  f.getValue().then(expectAsync1((v) {
    expect(v, equals([1, 2, 3, 4]));  
    expect(f.name, equals(const Symbol('li')));
    expect(f.type, equals(List));
    expect(f.isReadOnly, isFalse);
    expect(f.children, isEmpty);
  }));
}

@Test('Test the construction of a map field.')
void testMap() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('m'), Map);    
  f.getValue().then(expectAsync1((v) {
    expect(v, equals({ 'a' : 1, 'b' : 2 }));  
    expect(f.name, equals(const Symbol('m')));
    expect(f.type, equals(Map));
    expect(f.isReadOnly, isFalse);
    expect(f.children, isEmpty);
  }));
}

@Test('Test that setting a value on a read-only field throws.')
void testSetValueReadOnlyThrows() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('ri'), int, isReadOnly: true);
  f.setValue(13).catchError(expectAsync1((e) {
    expect(e, isStateError);
  }));
}

@Test('Test that setting a value adds a new data event to the stream.')
void testSetValue() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('i'), int);

  int dataCount = 0;  
  f.values.listen(expectAsync1((v) {
    expect(v, equals(8));
  }));
  
  f.setValue(8).then(expectAsync1((v) {
    expect(v, equals(8));
  }));
}

@Test('Test that no events are added to the values stream when setting a new'
      ' value if there are no listeners.')
void testSetValueNoValuesListener() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('i'), int);
  
  f.setValue(8).then(expectAsync1((v) {
    expect(v, equals(8));
    final noop = expectAsync0(() {});
    
    // Now attach a listener.
    f.values.listen((_) => fail('No values should be buffered.'));
    
    // Delay test completion to make sure no events are flushed to listener.
    runAsync(noop);
  }));
}

// -----------------------------------------------------------------------------
// Custom
// -----------------------------------------------------------------------------

@Test('Test the construction of a custom type field.')
void testCustom() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('c'), CustomType);    
  f.getValue().then(expectAsync1((v) {
    expect(v, const isInstanceOf<CustomType>());  
    expect(f.name, equals(const Symbol('c')));
    expect(f.type, equals(CustomType));
    expect(f.isReadOnly, isFalse);
    expect(f.children.length, equals(3));
  }));
}

@Test('Test that a child set value affects parent + sibling values.')
void testSetChildValue() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final f = new Field(mirror, const Symbol('c'), CustomType);
  
  // Set value listeners on the children
  final i = f.children.firstWhere((c) => c.name == const Symbol('i'));
  i.values.listen(expectAsync1((v) => expect(v, equals(9))));
  
  final ci = f.children.firstWhere((c) => c.name == const Symbol('ci'));
  ci.values.listen(expectAsync1((v) => expect(v, equals(9))));
  
  final s = f.children.firstWhere((c) => c.name == const Symbol('s'));
  s.values.listen((v) => fail('s did not change, expect no stream data.'));

  // Set value listener on the parent
  f.values.listen(expectAsync1((v) => 
      expect('$v', equals('custom(9,9,null)'))));
  
  // Set the value of 'i'
  i.setValue(9);
}
