
library fields.test.field_transform;

import 'package:bench/bench.dart';
import 'package:unittest/unittest.dart';
import '../lib/src/default_transforms.dart';
import '../lib/src/field_transform.dart';

main() => reflectTests();

// -----------------------------------------------------------------------------
// Setup
// -----------------------------------------------------------------------------

@Setup
void setup() => registerDefaultFieldTransforms();

@Teardown
void teardown() => FieldTransform.unregisterAll();

class CustomStringToIntTransform extends FieldTransform<String, int> {
  const CustomStringToIntTransform();
  call(value) => throw new UnimplementedError();
}

class Rectangle {
  int x, y, w, h;
}

class RectangleF {
  double x, y, w, h;
}

class RectangleToRectangleFTransform extends FieldTransform<Rectangle, RectangleF> {
  const RectangleToRectangleFTransform();
  call(value) => new RectangleF()..x = value.x.toDouble()
      ..y = value.y.toDouble()
      ..w = value.w.toDouble()
      ..h = value.h.toDouble();
}

// -----------------------------------------------------------------------------
// Basic
// -----------------------------------------------------------------------------

@Test('Test that simple types can be transformed.')
void testCanTransformSimpleTypes() {
  expect(canTransformField(int, int), isTrue);
  expect(canTransformField(String, int), isTrue);
}

@Test('Test a valid string to int transform.')
void testTransformStringToInt() {
  final String s = '42';
  final int i = transformField(s, String, int);
  expect(i, equals(42));
}

@Test('Test an erroneous string to int transform.')
void testTransformStringToIntThrows() {
  final String s = '42.123';
  expect(() {
    transformField(s, String, int);
  }, throwsA(const isInstanceOf<FieldTransformError>()));
}

@Test('Test that a string to int transform where value is! a string throws.')
void testTransformStringToIntValueIsNotStringThrows() {
  final map = { 'foo' : 'bar' };
  expect(() { 
    transformField(map, String, int); 
  }, throwsA(const isInstanceOf<TypeError>()));
}

// -----------------------------------------------------------------------------
// Custom
// -----------------------------------------------------------------------------

@Test('Test the registration of a custom string to int transform.')
void testRegisterCustomTransformStringToInt() {
  FieldTransform.register(const CustomStringToIntTransform());  
  expect(canTransformField(String, int), isTrue);
}

@Test('Test the registration of a custom rectangle transform.')
void testRegisterRectangleTransform() {
  expect(canTransformField(Rectangle, RectangleF), isFalse);
  FieldTransform.register(const RectangleToRectangleFTransform());
  expect(canTransformField(Rectangle, RectangleF), isTrue);
  expect(FieldTransform.unregister(const RectangleToRectangleFTransform()), isTrue);
  expect(canTransformField(Rectangle, RectangleF), isFalse);
}

@Test('Test a valid rectangle transform.')
void testTransformRectangle() {  
  FieldTransform.register(const RectangleToRectangleFTransform());  
  final r = new Rectangle()..x = 1..y = 2..w = 3..h = 4;
  final rf = transformField(r, Rectangle, RectangleF);
  expect(rf, const isInstanceOf<RectangleF>());
  expect(rf.x, equals(1));
  expect(rf.y, equals(2));
  expect(rf.w, equals(3));
  expect(rf.h, equals(4));
  expect(FieldTransform.unregister(const RectangleToRectangleFTransform()), isTrue);
}
